

import grammar.Grammar;
import grammar.util.*;
import parser.*;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.IOException;

/**
 * Kooprey GUI.
 *
 * @author Mathias Ricken
 */
public class KoopreyFrame extends JFrame {
    /**
     * Program date.
     */
    private static final String VERSION_DATE = "November 2006";

    /**
     * Content pane.
     */
    private JPanel _contentPane;

    /**
     * File chooser.
     */
    private JFileChooser _fcInput;

    /**
     * File chooser.
     */
    private JFileChooser _fcOutput;

    /**
     * Output text area.
     */
    private JTextArea _outputTA = new JTextArea();

    /**
     * Top panel.
     */
    private JPanel _topPanel = new JPanel();

    /**
     * Bottom panel.
     */
    private JPanel _bottomPanel = new JPanel();

    /**
     * Bottom scroll pane.
     */
    private JScrollPane _scrollPane = new JScrollPane();

    /**
     * File name input field.
     */
    private JTextField _inpFileTF = new JTextField();

    /**
     * Output path input field.
     */
    private JTextField _outPathTF = new JTextField();

    /**
     * Output package input field.
     */
    private JTextField _packageTF = new JTextField();

    /**
     * Button to change the input file.
     */
    private JButton _inputBtn = new JButton();

    /**
     * Button to change the output file.
     */
    private JButton _outputBtn = new JButton();

    /**
     * Button to parse file.
     */
    private JButton _parseBtn = new JButton();

    /**
     * Button to left-factor.
     */
    private JButton _leftFactorBtn = new JButton();

    /**
     * Button to atomize.
     */
    private JButton _atomizeBtn = new JButton();

    /**
     * Button to generateHelper.
     */
    private JButton _generateBtn = new JButton();

    /**
     * Button to show about dialog.
     */
    private JButton _aboutBtn = new JButton();

    /**
     * Factory combo box.
     */
    private JComboBox _factoryBox = new JComboBox();

    private ILambda[] _factoryLambdas = new ILambda[] {
        new ILambda() {
            public Object apply(Object param) {
                try {
                    InnerClassesFactoryGenerator.Singleton.generate(_grammar.getDefinitions().get(0).getName(), _outPathTF.getText(), _packageTF.getText(), _grammar);
                }
                catch (FileNotFoundException e) {
                    throw new RuntimeException(e);
                }
                return null;
            }
        },
        new ILambda() {
            public Object apply(Object param) {
                for(String terminal: _grammar.getTerminals()) {
                    try {
                        FilesFactoryGenerator.Singleton.generate(terminal, _outPathTF.getText(), _packageTF.getText(), _grammar);
                    }
                    catch (FileNotFoundException e) {
                        throw new RuntimeException(e);
                    }
                }

                for(String nonTerminal: _grammar.getNonTerminals()) {
                    try {
                        FilesFactoryGenerator.Singleton.generate(nonTerminal, _outPathTF.getText(), _packageTF.getText(), _grammar);
                    }
                    catch (FileNotFoundException e) {
                        throw new RuntimeException(e);
                    }
                }
                return null;
            }
        }
    };

    private ILambda selectedFactoryLambda = _factoryLambdas[0];

    /**
     * Factory for token visitors to parse D non-terminals.
     */
    private ParseDFactory _dFac;

    /**
     * Factory for token visitors to parse E non-terminals.
     */
    private ParseEFactory _eFac;

    /**
     * Factory for token visitors to parse T non-terminals.
     */
    private ParseTFactory _tFac;

    /**
     * Tokenizer.
     */
    private ITokenizer _tok;

    /**
     * Grammar.
     */
    private Grammar _grammar;

    /**
     * Constructor for the frame.
     */
    public KoopreyFrame() {
        enableEvents(AWTEvent.WINDOW_EVENT_MASK);
        try {
            jbInit();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Initialize GUI components.
     */
    private void jbInit() {
        try {
            _fcInput = new JFileChooser(new File(".").getCanonicalPath());
        }
        catch(IOException ioe) {
            _fcInput = new JFileChooser();
        }
        _fcInput.setDialogTitle("Open input file");
        try {
            _fcOutput = new JFileChooser(new File(".").getCanonicalPath());
        }
        catch(IOException ioe) {
            _fcOutput = new JFileChooser();
        }
        _fcOutput.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        _fcOutput.setDialogTitle("Select output directory");

        _contentPane = (JPanel) this.getContentPane();
        _contentPane.setLayout(new BorderLayout());

        this.setSize(new Dimension(616, 317));
        this.setTitle("Kooprey");

        _inputBtn.setText("Input file:");
        _inputBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                inputBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_inputBtn, null);
        _inpFileTF.setMinimumSize(new Dimension(100, 21));
        _inpFileTF.setPreferredSize(new Dimension(100, 21));
        _inpFileTF.setText("");
        _inpFileTF.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                parseBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_inpFileTF, null);

        _parseBtn.setText("Parse");
        _parseBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                parseBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_parseBtn, null);

        _leftFactorBtn.setText("Left-Factor");
        _leftFactorBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                leftFactorBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_leftFactorBtn, null);

        _atomizeBtn.setText("Atomize");
        _atomizeBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                atomizeBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_atomizeBtn, null);

        _aboutBtn.setText("About");
        _aboutBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                aboutBtn_actionPerformed(e);
            }
        });
        _topPanel.add(_aboutBtn, null);

        _bottomPanel.add(new JLabel("Factories as:"));
        _factoryBox.setModel(new DefaultComboBoxModel(new String[] {"inner classes", "separate files"}));
        _factoryBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                factoryBox_actionPerformed(e);
            }
        });
        _bottomPanel.add(_factoryBox, null);

        _outputBtn.setText("Output directory:");
        _outputBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                outputBtn_actionPerformed(e);
            }
        });
        _bottomPanel.add(_outputBtn, null);
        _outPathTF.setMinimumSize(new Dimension(100, 21));
        _outPathTF.setPreferredSize(new Dimension(100, 21));
        _outPathTF.setText("");
        _outPathTF.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                generateBtn_actionPerformed(e);
            }
        });
        _bottomPanel.add(_outPathTF, null);

        _bottomPanel.add(new JLabel("Package name:"));
        _packageTF.setMinimumSize(new Dimension(100, 21));
        _packageTF.setPreferredSize(new Dimension(100, 21));
        _packageTF.setText("");
        _packageTF.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                generateBtn_actionPerformed(e);
            }
        });
        _bottomPanel.add(_packageTF, null);

        _generateBtn.setText("Generate");
        _generateBtn.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                generateBtn_actionPerformed(e);
            }
        });
        _bottomPanel.add(_generateBtn, null);

        _contentPane.add(_topPanel, BorderLayout.NORTH);
        _contentPane.add(_scrollPane, BorderLayout.CENTER);
        _contentPane.add(_bottomPanel, BorderLayout.SOUTH);
        _scrollPane.getViewport().add(_outputTA, null);
        _scrollPane.setMinimumSize(new Dimension(600, 280));
        _scrollPane.setPreferredSize(new Dimension(600, 280));
        pack();
    }

    /**
     * Overridden so we can exit when window is closed.
     *
     * @param e window event
     */
    protected void processWindowEvent(WindowEvent e) {
        super.processWindowEvent(e);
        if (e.getID() == WindowEvent.WINDOW_CLOSING) {
            System.exit(0);
        }
    }

    /**
     * Make parser.
     *
     * @throws FileNotFoundException thrown if file is not found
     */
    protected void makeParser() throws FileNotFoundException {
        _tok = new Tokenizer(_inpFileTF.getText());

        ParseEmptyFactory emptyFactory = new ParseEmptyFactory(_tok);
        ParseWordFactory wordFac = new ParseWordFactory(_tok);
        ParseQuotedStringFactory qsFac = new ParseQuotedStringFactory(_tok);

        ParseD1aFactory d1aFac = new ParseD1aFactory(_tok);
        ParseD1Factory d1Fac = new ParseD1Factory(_tok, d1aFac, emptyFactory);

        ParseE1aFactory e1aFac = new ParseE1aFactory(_tok);
        ParseE1Factory e1Fac = new ParseE1Factory(_tok, e1aFac, emptyFactory);

        ParseT3Factory t3Fac = new ParseT3Factory(_tok, emptyFactory);
        ParseT1Factory t1Fac = new ParseT1Factory(_tok);
        ParseT2Factory t2Fac = new ParseT2Factory(_tok, qsFac);
        _tFac = new ParseTFactory(_tok,
                                 t1Fac,
                                 t2Fac);

        _eFac = new ParseEFactory(_tok, _tFac, e1Fac);

        _dFac = new ParseDFactory(_tok, _eFac, d1Fac);

        t3Fac.setParseTFactory(_tFac);
        t1Fac.setParseT3Factory(t3Fac);
        t2Fac.setParseT3Factory(t3Fac);
        e1aFac.setParseEFactory(_eFac);
        d1aFac.setParseDFactory(_dFac);
    }

    /**
     * Parse file.
     *
     * @param e action event
     */
    void parseBtn_actionPerformed(ActionEvent e) {
        try {
            if (_inpFileTF.getText().length()==0) {
                if (!inputBtn_actionPerformed(null)) {
                    return;
                }
            }
            makeGrammar();

            _outputTA.setText(_grammar.toString());
            _outputTA.setCaretPosition(0);
            System.out.println("Grammar = " + _grammar);
        }
        catch (Exception e1) {
            StringBuffer sb = new StringBuffer();
            sb.append(e1.toString());
            sb.append('\n');
            StackTraceElement[] ste = e1.getStackTrace();
            for (int i = 0; i < ste.length; ++i) {
                sb.append('\t');
                sb.append(ste[i].toString());
                sb.append('\n');
            }
            _outputTA.setText(sb.toString());
            _outputTA.setCaretPosition(0);
            e1.printStackTrace();
        }
    }

    /**
     * Select input file.
     *
     * @param e action event
     */
    boolean inputBtn_actionPerformed(ActionEvent e) {
        final FileFilter ff = new FileFilter() {
            public boolean accept(File f) {
                return (f.getName().endsWith(".txt")) || (f.isDirectory());
            }

            public String getDescription() {
                return "*.txt";
            }
        };
        _fcInput.addChoosableFileFilter(ff);
        int returnVal = _fcInput.showOpenDialog(this);
        boolean retval = false;
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            _inpFileTF.setText(_fcInput.getSelectedFile().toString());
            retval = true;
        }
        _fcInput.removeChoosableFileFilter(ff);
        return retval;
    }

    /**
     * Select output file.
     *
     * @param e action event
     */
    boolean outputBtn_actionPerformed(ActionEvent e) {
        int returnVal = _fcOutput.showOpenDialog(this);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            _outPathTF.setText(_fcOutput.getSelectedFile().toString());
            return true;
        }
        else {
            return false;
        }
    }

    /**
     * Parse the BNF and make the grammar.
     * @return grammar object representing the parsed BNF
     * @throws FileNotFoundException
     */
    Grammar makeGrammar() throws FileNotFoundException {
        makeParser();

        IVisitDefault parser = _dFac.makeVisitor();

        D result = (D)_tok.getNextToken().execute(parser, null);

        _grammar = new Grammar(result);
        return _grammar;
    }

    /**
     * Left-Factor Grammar.
     *
     * @param e action event
     */
    void leftFactorBtn_actionPerformed(ActionEvent e) {
        try {
            Grammar g = _grammar.leftFactor();
            _outputTA.setText(g.toString() +
                             System.getProperty("line.separator") +
                             System.getProperty("line.separator") +
                             "Previous result:" +
                             System.getProperty("line.separator") +
                             _outputTA.getText());
            _outputTA.setCaretPosition(0);
            System.out.println("Left-factored grammar = " + g);
            _grammar = g;
        }
        catch (Exception e1) {
            StringBuffer sb = new StringBuffer();
            sb.append(e1.toString());
            sb.append('\n');
            StackTraceElement[] ste = e1.getStackTrace();
            for (int i = 0; i < ste.length; ++i) {
                sb.append('\t');
                sb.append(ste[i].toString());
                sb.append('\n');
            }
            _outputTA.setText(sb.toString());
            _outputTA.setCaretPosition(0);
            e1.printStackTrace();
        }
    }

    /**
     * Atomize Grammar.
     *
     * @param e action event
     */
    void atomizeBtn_actionPerformed(ActionEvent e) {
        try {
            Grammar g = _grammar.atomize();
            _outputTA.setText(g.toString() +
                             System.getProperty("line.separator") +
                             System.getProperty("line.separator") +
                             "Previous result:" +
                             System.getProperty("line.separator") +
                             _outputTA.getText());
            _outputTA.setCaretPosition(0);
            System.out.println("Left-factored grammar = " + g);
            _grammar = g;
        }
        catch (Exception e1) {
            StringBuffer sb = new StringBuffer();
            sb.append(e1.toString());
            sb.append('\n');
            StackTraceElement[] ste = e1.getStackTrace();
            for (int i = 0; i < ste.length; ++i) {
                sb.append('\t');
                sb.append(ste[i].toString());
                sb.append('\n');
            }
            _outputTA.setText(sb.toString());
            _outputTA.setCaretPosition(0);
            e1.printStackTrace();
        }
    }


    /**
     * Generate Files.
     *
     * @param e action event
     */
    void generateBtn_actionPerformed(ActionEvent e) {
        try {
            if (_inpFileTF.getText().length()==0) {
                if (!inputBtn_actionPerformed(null)) {
                    return;
                }
            }
            if (_outPathTF.getText().length()==0) {
                if (!outputBtn_actionPerformed(e)) {
                    return;
                }
            }
            _grammar = makeGrammar().leftFactor().atomize();
            _grammar.generate(_outPathTF.getText(), _packageTF.getText());

            ATokenGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            ATVFactoryGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            EOFTokenGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            ITokenizerGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            ITVFactoryGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            ITokVisitorGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            LambdaGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);
            TokenizerGenerator.Singleton.generate(_outPathTF.getText(), _packageTF.getText(), _grammar);

            for(String terminal: _grammar.getTerminals()) {
                TerminalGenerator.Singleton.generate(terminal, _outPathTF.getText(), _packageTF.getText(), _grammar);
            }

            selectedFactoryLambda.apply(null);
        }
        catch (Exception e1) {
            StringBuffer sb = new StringBuffer();
            sb.append(e1.toString());
            sb.append('\n');
            StackTraceElement[] ste = e1.getStackTrace();
            for (int i = 0; i < ste.length; ++i) {
                sb.append('\t');
                sb.append(ste[i].toString());
                sb.append('\n');
            }
            _outputTA.setText(sb.toString());
            _outputTA.setCaretPosition(0);
            e1.printStackTrace();
        }
    }

    /**
     * Brings up a simple dialog with some general information.
     */
    private void aboutBtn_actionPerformed(ActionEvent e) {
        String html =
                "<html><h2>Kooprey</h2>" +
                "A tool converting LL(1) BNF grammars to<p>" +
                "object-oriented parsers in Java.<p><p>" +
                "<font size=-1>Version " + VERSION_DATE + " written by Mathias Ricken.</font><p>" +
                "<font size=-1>Copyright&copy; 2004-2006 Rice University (www.rice.edu).<p></html>";
        JOptionPane.showMessageDialog(this, new JLabel(html), "About Kooprey", JOptionPane.INFORMATION_MESSAGE, null);
    }


    /**
     * Changes the way factories are generated.
     */
    private void factoryBox_actionPerformed(ActionEvent e) {
        selectedFactoryLambda = _factoryLambdas[_factoryBox.getSelectedIndex()];
    }
}