package grammar;

import parser.*;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;

/**
 * BNF Definition.
 *
 * @author Mathias Ricken
 */
public class Definition {
    /**
     * Name of this definition.
     */
    private String _name;

    /**
     * List of alternative sequences.
     */
    private ArrayList<Sequence> _alternatives = new ArrayList<Sequence>();

    /**
     * Returns a string representation of the object.
     *
     * @return a string representation of the object.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(_name);
        sb.append(" ::=");
        boolean first = true;
        for(Sequence s: _alternatives) {
            if (first) {
                first = false;
            }
            else {
                sb.append(" |");
            }
            sb.append(" " + s.toString());
        }
        return sb.toString();
    }

    /**
     * Create a definition from a symbol name and a list of alternative sequences.
     * @param name
     * @param alternatives
     */
    public Definition(String name, ArrayList<Sequence> alternatives) {
        _name = new String(name);
        _alternatives = new ArrayList<Sequence>();
        for(Sequence seq: alternatives) {
            _alternatives.add(new Sequence(seq));
        }
    }

    /**
     * Copy constructor.
     *
     * @param d
     */
    public Definition(Definition d) {
        this(d._name, d._alternatives);
    }

    /**
     * Create a definition from a symbol name and an expression.
     * @param symbol symbol name
     * @param e E
     */
    public Definition(WordToken symbol, E e) {
        _name = symbol.getLexeme();
        _alternatives.add(new Sequence(e.getSequence()));
        e.getRest().execute(new E1.IE1Visitor<Object, ArrayList<Sequence>>() {
            public Object e1aCase(E1a host, ArrayList<Sequence> sequences) {
                sequences.add(new Sequence(host.getExpression().getSequence()));
                return host.getExpression().getRest().execute(this, sequences);
            }
            public Object emptyCase(Empty host, ArrayList<Sequence> sequences) {
                return null;
            }
        }, _alternatives);

    }

    /**
     * Accessor for the name.
     * @return name
     */
    public String getName() {
        return _name;
    }

    /**
     * Mutator for the name.
     * @param name new name
     */
    public void setName(String name) {
        _name = name;
    }

    /**
     * Accessor for the alternatives.
     * @return alternatives
     */
    public ArrayList<Sequence> getAlternatives() {
        return _alternatives;
    }

    /**
     * Mutatorfor the alternatives.
     * @param alternatives new alternatives
     */
    public void setAlternatives(ArrayList<Sequence> alternatives) {
        _alternatives = alternatives;
    }

    /**
     * Return a list of all those sequences that start with the specified symbols.
     * @param sequences list of sequences
     * @param symbols start symbols
     * @return list of all those sequences that start with the specified symbol
     */
    protected ArrayList<Sequence> getSequencesStartingWith(ArrayList<Sequence> sequences, ArrayList<String> symbols) {
        ArrayList<Sequence> seqs = new ArrayList<Sequence>();
        Sequence startSeq = new Sequence(symbols);

        for(Sequence s: sequences) {
            if (s.toString().startsWith(startSeq.toString())) {
                seqs.add(s);
            }
        }

        return seqs;
    }

    /**
     * Left-factor this definition.
     * @param g grammar that contains this definition
     * @return true if left-factoring occured was done
     */
    public boolean leftFactor(Grammar g) {
        if (getAlternatives().size()<2) {
            // less than two alternatives, cannot left-factor
            return false;
        }

        boolean didSomething = false;
        boolean done;

        do {
            done = true;

            for (Sequence s: _alternatives) {
                int i = 0;
                ArrayList<String> newStartSymbols = new ArrayList<String>();
                newStartSymbols.add(s.getSymbols().get(i));
                ArrayList<Sequence> newStartWithSeqs = getSequencesStartingWith(_alternatives, newStartSymbols);

                if (newStartWithSeqs.size()>1) {
                    // yes, we need to left-factor
                    didSomething = true;
                    done = false;

                    // find the shortest sequence
                    int shortestLength = Integer.MAX_VALUE;
                    for(Sequence factorSeq: newStartWithSeqs) {
                        if (shortestLength>factorSeq.getSymbols().size()) {
                            shortestLength = factorSeq.getSymbols().size();
                        }
                    }

                    // see if we can add more start symbols without shrinking the set
                    ArrayList<String> startSymbols;
                    ArrayList<Sequence> startWithSeqs;
                    if (shortestLength>1) {
                        do {
                            // save old values
                            ++i;
                            startSymbols = new ArrayList<String>(newStartSymbols);
                            startWithSeqs = newStartWithSeqs;

                            // create new
                            newStartSymbols.add(s.getSymbols().get(i));
                            newStartWithSeqs = getSequencesStartingWith(_alternatives, newStartSymbols);
                        } while ((newStartWithSeqs.size() == startWithSeqs.size()) && (i+1 < shortestLength));

                        if (newStartWithSeqs.size() == startWithSeqs.size()) {
                            // aborted because we can't grow it longer than the shortest sequence
                            // even though this still matched
                            startSymbols = newStartSymbols;
                            startWithSeqs = newStartWithSeqs;
                        }
                    }
                    else {
                        // no, cannot grow
                        // shortest sequence is only one symbol
                        startSymbols = newStartSymbols;
                        startWithSeqs = newStartWithSeqs;
                    }

                    // we have the start symbols in startSymbols and the list of sequences
                    // starting with those in startWithSeqs

                    // System.out.println("Left factoring "+toString());

                    Sequence replacementSequence = new Sequence(startSymbols);

                    // System.out.println("Factor = "+(replacementSequence.toString()));
                    for(Sequence productSeq: startWithSeqs) {
                        // remove this sequence from the original definition
                        _alternatives.remove(productSeq);

                        // remove the factored symbols
                        for (int j=0; j<startSymbols.size(); ++j) {
                            productSeq.getSymbols().remove(0);
                        }
                        if (productSeq.getSymbols().size()==0) {
                            // sequence became empty, add "Empty" terminal
                            productSeq.getSymbols().add("Empty");
                        }
                        // System.out.println("\t" + productSeq);
                    }

                    String nextName = getNextName(g);

                    replacementSequence.getSymbols().add(nextName);
                    _alternatives.add(replacementSequence);
                    // System.out.println("Definition = "+toString());

                    Definition d = new Definition(nextName, startWithSeqs);

                    g.getDefinitions().add(d);

                    break;
                }
            }
        } while(!done);

        return didSomething;
    }

    /**
     * Get the next free name.
     * @param g grammar
     * @return next free name
     */
    protected String getNextName(Grammar g) {
        int largest = 0;
        for (Definition d: g.getDefinitions()) {
            String n = d.getName();
            if (n.startsWith(_name+"_")) {
                // starts with "definitionName_"
                int npos = _name.length()+1;
                if (n.indexOf('_', npos)==-1) {
                    // does not have another underscore afterwards
                    String s = n.substring(npos);
                    try {
                        int i = Integer.parseInt(s);
                        if (i > largest) {
                            largest = i;
                        }
                    }
                    catch (NumberFormatException e) {
                        // wasn't a number, ignore
                    }
                }
            }
        }

        return _name + "_" + (largest+1);
    }

    /**
     * Modify this definition so that alternatives are atomic, i.e. that they can be modelled as interfaces.
     * @param g grammar that contains this definition
     * @return true if the definition was changed
     */
    public boolean atomize(Grammar g) {
        // System.out.println("Atomize "+toString());
        // System.out.println("\t"+getAlternatives().size()+" alternatives");

        if (getAlternatives().size()<2) {
            // less than two alternatives, cannot atomize
            return false;
        }

        boolean didSomething = false;

        for (Sequence s: _alternatives) {
            // atomize if we have:
            // a) a sequence of two or more symbols as alternative (e.g. A B | ...)
            // b) a single terminal, which is not Empty (e.g. Num | ...)
            if ((s.getSymbols().size()>1) || ((s.getSymbols().size()==1) && g.isTerminal(s.getSymbols().get(0))) && (!s.getSymbols().get(0).equals("Empty"))) {
                // System.out.println("\tAtomizing "+s);

                // yes, we need to atomize
                didSomething = true;

                String nextName = getNextName(g);

                ArrayList<Sequence> seq = new ArrayList<Sequence>();
                seq.add(s);
                Definition d = new Definition(nextName, seq);

                g.getDefinitions().add(d);

                s.getSymbols().clear();
                s.getSymbols().add(nextName);
            }
        }


        // move Empty alternatives to the end
        int i=0;
        boolean removedEmpty = false;
        while (i<_alternatives.size()) {
            Sequence s = _alternatives.get(i);
            if (s.getSymbols().get(0).equals("Empty")) {
                _alternatives.remove(s);
                removedEmpty = true;
            }
            else {
                ++i;
            }
        }
        if (removedEmpty) {
            ArrayList<String> emptyAlternative = new ArrayList<String>();
            emptyAlternative.add("Empty");
            _alternatives.add(new Sequence(emptyAlternative));
        }

        return didSomething;
    }

    /**
     * Return true if this is an alternative.
     * @return true if alternative
     */
    public boolean isAlternative() {
        return (getAlternatives().size()>1);
    }

    /**
     * Return true if this definition is either a pure sequence or made up of atomic alternatives.
     * @return true if pure sequence or all atomic alternatives
     */
    public boolean isAtomic() {
        if (!isAlternative()) {
            return true;
        }
        for (Sequence s: _alternatives) {
            if (s.getSymbols().size()>1) {
                return false;
            }
        }
        return true;
    }

    /**
     * Generate Java files for this definition.
     * @param path path to put the files
     * @param packageName package name
     * @param g grammar that contains this definition
     */
    public void generate(String path, String packageName, Grammar g) throws FileNotFoundException {
        PrintWriter pw = new PrintWriter(new FileOutputStream(path+File.separator+_name+".java"), true);

        if (!packageName.equals("")) {
            pw.println("package "+packageName+";");
            pw.println();
        }
        pw.println("/**");
        pw.println(" * Non-terminal: " + toString());
        pw.println(" */");

        if (isAlternative())  {
            generateAlternative(pw, path, packageName, g);
        }
        else {
            generateSequence(pw, path, packageName, g);
        }

        pw.println("}");

        pw.close();
    }

    /**
     * Generate Java files for this alternative.
     * @param pw PrintWriter
     * @param path path to put the files
     * @param packageName package name
     * @param g grammar that contains this definition
     */
    private void generateAlternative(PrintWriter pw, String path, String packageName, Grammar g) throws FileNotFoundException {
        // make sure this definition is atomized
        if (!isAlternative()) {
            throw new IllegalArgumentException("Definition has non-atomic alternatives");
        }

        pw.print("public interface " + _name);

        ArrayList<Definition> referers = g.getRefererDefinitions(_name);
        boolean first = true;

        for (Definition d: referers) {
            if (d.isAlternative()) {
                if (first) {
                    pw.println();
                    pw.print("    extends " + d.getName());
                    first = false;
                }
                else {
                    pw.println(",");
                    pw.print("            " + d.getName());
                }
            }
        }
        pw.println(" {");
    }

    /**
     * Generate Java files for this sequence.
     * @param pw PrintWriter
     * @param path path to put the files
     * @param packageName package name
     * @param g grammar that contains this definition
     */
    private void generateSequence(PrintWriter pw, String path, String packageName, Grammar g) throws FileNotFoundException {
        pw.print("public class " + _name);

        ArrayList<Definition> referers = g.getRefererDefinitions(_name);
        boolean first = true;

        for (Definition d: referers) {
            if (d.isAlternative()) {
                if (first) {
                    pw.println();
                    pw.print("    implements " + d.getName());
                    first = false;
                }
                else {
                    pw.println(",");
                    pw.print("               " + d.getName());
                }
            }
        }

        pw.println(" {");

        for (String symbol: _alternatives.get(0).getSymbols()) {
            if (!symbol.startsWith("\"")) {
                pw.println("    /**");
                if (g.isTerminal(symbol)) {
                    pw.println("     * Terminal: " + symbol + ".");
                }
                else {
                    pw.println("     * Non-terminal: " + symbol + ".");
                }
                pw.println("     */");
                pw.println("    private " + symbol + " " + symbolToVariableName(symbol) + ";");
                pw.println();
            }
        }

        pw.println("    /**");
        pw.println("     * Constructor for " + _name + ".");
        pw.println("     *");

        for (String symbol: _alternatives.get(0).getSymbols()) {
            if (symbol.startsWith("\"")) {
                symbol = quotedStringToName(g, symbol);

                pw.println("     * @param " + symbolToVariableName(symbol) + " Terminal " + symbol);
            }
            else {
                pw.print("     * @param " + symbolToVariableName(symbol) + " ");
                if (g.isTerminal(symbol)) {
                    pw.println("Terminal " + symbol);
                }
                else {
                    pw.println("Non-terminal " + symbol);
                }
            }
        }

        pw.println("     */");
        String signatureLine = "    public " + _name + "(";
        String pad = "";
        for (int i = 0; i < signatureLine.length(); ++i ) {
            pad = pad + " ";
        }
        pw.print(signatureLine);

        first = true;
        for (String symbol: _alternatives.get(0).getSymbols()) {
            if (symbol.startsWith("\"")) {
                symbol = quotedStringToName(g, symbol);
            }
            if (first) {
                first = false;
            }
            else {
                pw.println(",");
                pw.print(pad);
            }
            pw.print(symbol + " " + symbolToVariableName(symbol));
        }
        pw.println(") {");

        for (String symbol: _alternatives.get(0).getSymbols()) {
            if (!symbol.startsWith("\"")) {
                pw.println("        this." + symbolToVariableName(symbol) + " = " +  symbolToVariableName(symbol) + ";");
            }
        }

        pw.println("    }");

        pw.println();
        pw.println("    /**");
        pw.println("     * Return a string representation.");
        pw.println("     *");
        pw.println("     * @return string representation");
        pw.println("     */");
        pw.print("    public String toString() {");

        first = true;
        for (String symbol: _alternatives.get(0).getSymbols()) {
            pw.println();
            String output;
            if (symbol.startsWith("\"")) {
                output = symbol;
            }
            else {
                output = symbolToVariableName(symbol) + ".toString()";
            }
            if (first) {
                pw.print("        return " + output);
                first = false;
            }
            else {
                pw.print("             + " + output);
            }
        }
        pw.println(";");

        pw.println("    }");
    }

    protected String symbolToVariableName(String symbol) {
        return "_" + symbol.toLowerCase();
    }


    /**
     * Get the name from a quoted string.
     * @param g grammar
     * @param symbol quoted string
     * @return name
     */
    private String quotedStringToName(Grammar g, String symbol) {
        String name;
        name = "_Terminal" + g.getTerminals().indexOf(symbol);
        return name;
    }
}
