package grammar;

import parser.D;
import parser.D1;
import parser.D1a;
import parser.Empty;

import java.util.ArrayList;
import java.io.FileNotFoundException;

/**
 * Grammar.
 *
 * @author Mathias Ricken
 */
public class Grammar {
    /**
     * List of definitions.
     */
    private ArrayList<Definition> _definitions = new ArrayList<Definition>();

    /**
     * List of symbols presumed terminals.
     */
    private ArrayList<String> _terminals = new ArrayList<String>();

    /**
     * List of symbols that are non-terminals.
     */
    private ArrayList<String> _nonTerminals = new ArrayList<String>();

    /**
     * Create a grammar from a list of definitions.
     *
     * @param list
     */
    public Grammar(ArrayList<Definition> list) {
        _definitions = new ArrayList<Definition>();
        for(Definition def: list) {
            _definitions.add(new Definition(def));
        }

        buildSymbolLists();
    }

    /**
     * Copy constructor.
     *
     * @param g other grammar
     */
    public Grammar(Grammar g) {
        this(g._definitions);
    }

    /**
     * Create a grammar from a d.
     *
     * @param d D
     */
    public Grammar(D d) {
        _definitions.add(new Definition(d.getSymbol(), d.getExpression()));
        d.getRest().execute(new D1.ID1Visitor<Object, ArrayList<Definition>>() {
            public Object d1aCase(D1a host, ArrayList<Definition> definitions) {
                definitions.add(new Definition(host.getDefinition().getSymbol(), host.getDefinition().getExpression()));
                return host.getDefinition().getRest().execute(this, definitions);
            }
            public Object emptyCase(Empty host, ArrayList<Definition> definitions) {
                return null;
            }
        }, _definitions);

        buildSymbolLists();
    }

    /**
     * Get the definition for the specified symbol, or return null.
     * @param s symbol
     * @return definition or null
     */
    public Definition getDefinition(String s) {
        for(Definition d: _definitions) {
            if (d.getName().equals(s)) {
                return d;
            }
        }
        return null;
    }

    /**
     * Return a list of definitions that refer to this definition.
     * @return
     */
    public ArrayList<Definition> getRefererDefinitions(String symbol) {
        ArrayList<Definition> referers = new ArrayList<Definition>();
        for(Definition d: _definitions) {
            for (Sequence seq: d.getAlternatives()) {
                if (seq.getSymbols().contains(symbol)) {
                    referers.add(d);
                }
            }
        }

        return referers;
    }

    /**
     * Get the definitions.
     * @return definitions
     */
    public ArrayList<Definition> getDefinitions() {
        return _definitions;
    }

    /**
     * Fill lists of terminals and non-terminals.
     */
    public void buildSymbolLists() {
        _terminals = new ArrayList<String>();
        _nonTerminals = new ArrayList<String>();

        // fill list of non-terminals
        for(Definition def: _definitions) {
            _nonTerminals.add(def.getName());
        }

        // fill list of terminals
        for(Definition def: _definitions) {
            for (Sequence seq: def.getAlternatives()) {
                for (String sym: seq.getSymbols()) {
                    if ((!_nonTerminals.contains(sym)) && (!_terminals.contains(sym))) {
                        _terminals.add(sym);
                    }
                }
            }
        }
    }

    /**
     * Returns a string representation of the object.
     *
     * @return a string representation of the object.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        boolean first = true;
        for(Definition d: _definitions) {
            if (first) {
                first = false;
            }
            else {
                sb.append(";");
                sb.append(System.getProperty("line.separator"));
            }
            sb.append(d.toString());
        }
        sb.append(System.getProperty("line.separator"));
        sb.append("Terminals:");
        sb.append(System.getProperty("line.separator"));
        sb.append('\t');
        for(String terminal: _terminals) {
            sb.append(terminal);
            sb.append(" ");
        }

        return sb.toString();
    }

    /**
     * Create a left-factored grammar.
     * @return left-factored grammar
     */
    public Grammar leftFactor() {
        Grammar g = new Grammar(this);

        for (int defIndex=0; defIndex<g._definitions.size(); ++defIndex) {
            Definition d = g._definitions.get(defIndex);

            d.leftFactor(g);
        }

        g.buildSymbolLists();

        // maintain old order
        ArrayList<String> lfTerminals = new ArrayList<String>(g.getTerminals());
        g.getTerminals().clear();
        g.getTerminals().addAll(_terminals);
        for(String terminal: lfTerminals) {
            if(!g.getTerminals().contains(terminal)) {
                g.getTerminals().add(terminal);
            }
        }
        return g;
    }

    /**
     * Create a grammar with atomic alternatives, i.e. alternatives that can be modeled as interfaces.
     * @return grammar with interface alternatives
     */
    public Grammar atomize() {
        Grammar g = new Grammar(this);

        for (int defIndex=0; defIndex<g._definitions.size(); ++defIndex) {
            Definition d = g._definitions.get(defIndex);

            d.atomize(g);
        }

        g.buildSymbolLists();

        // maintain old order
        ArrayList<String> atTerminals = new ArrayList<String>(g.getTerminals());
        g.getTerminals().clear();
        g.getTerminals().addAll(_terminals);
        for(String terminal: atTerminals) {
            if(!g.getTerminals().contains(terminal)) {
                g.getTerminals().add(terminal);
            }
        }

        return g;
    }

    /**
     * Generate Java files for this grammar.
     * @param path path to put the files
     * @param packageName package name
     */
    public void generate(String path, String packageName) throws FileNotFoundException {
        for (Definition d: _definitions) {
            d.generate(path, packageName, this);
        }
    }

    /**
     * Returns true if this symbol is a terminal.
     * @param symbol
     * @return true if terminal
     */
    public boolean isTerminal(String symbol) {
        return (_terminals.contains(symbol));
    }

    /**
     * Accessor for the list of terminals.
     * @return list of terminals
     */
    public ArrayList<String> getTerminals() {
        return _terminals;
    }

    /**
     * Accessor for the list of non-terminals.
     * @return list of non-terminals
     */
    public ArrayList<String> getNonTerminals() {
        return _nonTerminals;
    }
}
