package grammar.util;

import grammar.Grammar;
import grammar.Definition;
import grammar.Sequence;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;

/**
 * Generator for factories.
 *
 * @author Mathias Ricken
 */
public class FactoryGenerator {
    /**
     * Singleton instance.
     */
    public static final FactoryGenerator Singleton = new FactoryGenerator();

    /**
     * Private singleton constructor.
     */
    private FactoryGenerator() {
        // nothing to do
    }

    private static void println(PrintWriter pw, int indent, String s) {
        for (int i=0; i<indent; ++i) {
            pw.print("    ");
        }
        pw.println(s);
    }

    private static void print(PrintWriter pw, int indent, String s) {
        for (int i=0; i<indent; ++i) {
            pw.print("    ");
        }
        pw.print(s);
    }

    public void generate(PrintWriter pw, int indent, String symbol, String path, String packageName, Grammar g, boolean ctorParams, ArrayList<String> symbolStack, ILambda accessSpecifier) throws FileNotFoundException {
        String name = symbol;
        if (symbol.equals("Empty")) {
            EmptyFactGenerator.Singleton.generate(pw, indent, path, packageName, g, accessSpecifier);
            return;
        }

        if (symbol.startsWith("\"")) {
            name = quotedStringToName(g, symbol);
        }

        println(pw, indent, "/**");
        println(pw, indent, " * Factory for " + name + ".");
        println(pw, indent, " */");
        println(pw, indent, accessSpecifier.apply(symbol) + " class " + name + "Fact extends ATVFactory {");

        Definition def = g.getDefinition(symbol);

        if (def==null) {
            generateTerminal(pw, indent, g, name);
        }
        else {
            generateNonTerminal(pw, indent, def, g, name, ctorParams, symbolStack);
        }
    }

    private void generateTerminal(PrintWriter pw, int indent, Grammar g, String name) {
        println(pw, indent, "    /**");
        println(pw, indent, "     * Constructor for the " + name + " factory.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @param tkz   tokenizer to use");
        println(pw, indent, "     */");
        println(pw, indent, "    public " + name + "Fact(ITokenizer tkz) {");
        println(pw, indent, "        super(tkz);");
        println(pw, indent, "    }");
        println(pw, indent, "");
        println(pw, indent, "");
        println(pw, indent, "    /**");
        println(pw, indent, "     * Make a token visitor to parse an E non-terminal.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @return token visitor");
        println(pw, indent, "     */");
        println(pw, indent, "    public ITokVisitor makeVisitor() {");
        println(pw, indent, "        return new " + name + ".I" + name + "Visitor() {");
        println(pw, indent, "            public Object " + symbolToCaseName(name) + "(" + name + " host, Object param) {");
        println(pw, indent, "                return host;");
        println(pw, indent, "            }");
        println(pw, indent, "");
        println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
        println(pw, indent, "                throw new IllegalArgumentException(\"Wrong token: '\" + host + \"'\");");
        println(pw, indent, "            }");
        println(pw, indent, "        };");
        println(pw, indent, "    }");
        println(pw, indent, "");
        println(pw, indent, "    /**");
        println(pw, indent, "     * Make a token visitor that delegates to the given visitor in a chain of responsibility.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @param successor visitor to serve as successor in the chain");
        println(pw, indent, "     */");
        println(pw, indent, "    public ITokVisitor makeChainedVisitor(final ITokVisitor successor) {");
        println(pw, indent, "        return new " + name + ".I" + name + "Visitor() {");
        println(pw, indent, "            public Object " + symbolToCaseName(name) + "(" + name + " host, Object param) {");
        println(pw, indent, "                return host;");
        println(pw, indent, "            }");
        println(pw, indent, "");
        println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
        println(pw, indent, "                return host.execute(successor, param);");
        println(pw, indent, "            }");
        println(pw, indent, "        };");
        println(pw, indent, "    }");
    }

    private void generateNonTerminal(PrintWriter pw, int indent, Definition def, Grammar g, String name, boolean ctorParams, ArrayList<String> symbolStack) {
        ArrayList<String> referencedSymbols = new ArrayList<String>();
        ArrayList<String> origReferencedSymbols = new ArrayList<String>();

        if (def.isAlternative()) {
            for (Sequence seq: def.getAlternatives()) {
                String ref = seq.getSymbols().get(0);
                origReferencedSymbols.add(ref);
                if (ref.startsWith("\"")) {
                    ref = quotedStringToName(g, ref);
                }
                referencedSymbols.add(ref);
            }
        }
        else {
            for (String sym: def.getAlternatives().get(0).getSymbols()) {
                origReferencedSymbols.add(sym);
                if (sym.startsWith("\"")) {
                    sym = quotedStringToName(g, sym);
                }
                referencedSymbols.add(sym);
            }

        }

        for(String ref: referencedSymbols) {
            println(pw, indent, "    /**");
            println(pw, indent, "     * Visitor factory to parse " + ref + ".");
            println(pw, indent, "     */");
            println(pw, indent, "    private " + ref + "Fact " + symbolToVariableName(ref+"Fact") + ";");
            if (!def.isAlternative()) {
                println(pw, indent, "");
                println(pw, indent, "    /**");
                println(pw, indent, "     * Token visitor to parse " + ref + ".");
                println(pw, indent, "     */");
                println(pw, indent, "    private ITokVisitor " + symbolToVariableName("parse"+ref) + ";");
            }
            println(pw, indent, "");
        }

        if (!def.isAlternative()) {

            println(pw, indent, "    /**");
            println(pw, indent, "     * Initializer lambda for this factory.");
            println(pw, indent, "     */");
            println(pw, indent, "    private ILambda _initializer = new ILambda() {");
            println(pw, indent, "        public Object apply(Object param) {");
            println(pw, indent, "            // change state to no-op");
            println(pw, indent, "            _initializer = NoOpLambda.Singleton;");
            println(pw, indent, "");
            println(pw, indent, "            // initialize");

            for (String ref: referencedSymbols) {
                println(pw, indent, "            " + symbolToVariableName("parse"+ref)+ " = " + symbolToVariableName(ref+"Fact") + ".makeVisitor();");
            }

            println(pw, indent, "            return null;");
            println(pw, indent, "        }");
            println(pw, indent, "    };");

        }

        println(pw, indent, "    /**");
        println(pw, indent, "     * Constructor for the " + name + " factory.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @param tkz   tokenizer to use");

        if (ctorParams) {
            for(String ref: referencedSymbols) {
                println(pw, indent, "     * @param " + symbolToVariableName(ref+"Fact") + " factory for " + ref + " to use");
            }
        }

        println(pw, indent, "     */");

        String signatureLine = "    public " + name + "Fact(";
        String pad = "";
        for (int i = 0; i < signatureLine.length(); ++i ) {
            pad = pad + " ";
        }
        print(pw, indent, signatureLine+"ITokenizer tkz");

        if (ctorParams) {
            for (String ref: referencedSymbols) {
                pw.println(",");
                print(pw, indent, pad+ref+"Fact "+symbolToVariableName(ref+"Fact"));
            }
        }
        pw.println(") {");
        println(pw, indent, "        super(tkz);");

        if (ctorParams) {
            for (String ref: referencedSymbols) {
                println(pw, indent, "        this." + symbolToVariableName(ref+"Fact") + " = " + symbolToVariableName(ref+"Fact") + ";");
            }
        }
        else {
            for (String ref: referencedSymbols) {
                if (symbolStack.contains(ref)) {
                    println(pw, indent, "        this." + symbolToVariableName(ref+"Fact") + " = " + ref + "Fact.this;");
                }
                else {
                    println(pw, indent, "        this." + symbolToVariableName(ref+"Fact") + " = new " + ref + "Fact(tkz);");
                }
            }
        }

        println(pw, indent, "    }");
        println(pw, indent, "");

        if (ctorParams) {
            println(pw, indent, "    /**");
            println(pw, indent, "     * Constructor for the " + name + " factory.");
            println(pw, indent, "     * This does not set the factories, they have to be set using the setXXXFact() methods.");
            println(pw, indent, "     *");
            println(pw, indent, "     * @param tkz   tokenizer to use");
            println(pw, indent, "     */");
            println(pw, indent, "    public " + name + "Fact(ITokenizer tkz) {");
            println(pw, indent, "        super(tkz);");
            println(pw, indent, "    }");
            println(pw, indent, "");

            for (String ref: referencedSymbols) {
                println(pw, indent, "    /**");
                println(pw, indent, "     * Mutator for the " + ref + " factory.");
                println(pw, indent, "     *");
                println(pw, indent, "     * @param " + symbolToVariableName(ref+"Fact") + " factory for " + ref + " to use");
                println(pw, indent, "     */");
                println(pw, indent, "    public void set" + ref + "Fact("+ref+"Fact "+symbolToVariableName(ref+"Fact")+") {");
                println(pw, indent, "        this." + symbolToVariableName(ref+"Fact") + " = " +  symbolToVariableName(ref+"Fact") + ";");
                println(pw, indent, "    }");
                println(pw, indent, "");
            }
        }

        if (!def.isAlternative()) {
            println(pw, indent, "    /**");
            println(pw, indent, "     * Make a " + name + ".");
            println(pw, indent, "     *");
            println(pw, indent, "     * @param host some token");
            println(pw, indent, "     * @param param visitor-specific input");
            println(pw, indent, "     * @return D grammar non-terminal");
            println(pw, indent, "     */");
            println(pw, indent, "    private " + name + " make" + name + "(AToken host, Object param) {");
            String returnLine = "        return new " + name + "(";
            print(pw, indent, returnLine);
            pad = "";
            for (int i = 0; i < returnLine.length(); ++i ) {
                pad = pad + " ";
            }
            boolean first = true;
            for (String ref: referencedSymbols) {
                if (first) {
                    first = false;
                    pw.print("(" + ref + ") host");
                }
                else {
                    pw.println(",");
                    print(pw, indent, pad);
                    pw.print("(" + ref + ") nextToken()");
                }
                pw.print(".execute("+symbolToVariableName("parse"+ref)+", param)");
            }
            pw.println(");");
            println(pw, indent, "    }");
            println(pw, indent, "");
        }
        println(pw, indent, "    /**");
        println(pw, indent, "     * Make a token visitor to parse an E non-terminal.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @return token visitor");
        println(pw, indent, "     */");
        println(pw, indent, "    public ITokVisitor makeVisitor() {");
        if (def.isAlternative()) {
            print(pw, indent, "        return ");
            for (int i=0; i < referencedSymbols.size()-1; ++i) {
                String ref = referencedSymbols.get(i);
                pw.println(symbolToVariableName(ref+"Fact") + ".makeChainedVisitor(");
                print(pw, indent, "               ");
            }
            String ref = referencedSymbols.get(referencedSymbols.size()-1);
            pw.print(symbolToVariableName(ref+"Fact") + ".makeVisitor(");
            for (int i=0; i < referencedSymbols.size(); ++i) {
                pw.print(")");
            }
            pw.println(";");
        }
        else {
            println(pw, indent, "        initialize();");
            if (g.isTerminal(origReferencedSymbols.get(0))) {
                // first one is terminal
                println(pw, indent, "        return new " + referencedSymbols.get(0) + ".I" + referencedSymbols.get(0) + "Visitor() {");
                println(pw, indent, "            public Object " + symbolToCaseName(referencedSymbols.get(0)) + "(" + referencedSymbols.get(0) + " host, Object param) {");
                println(pw, indent, "                return make" + name + "(host, param);");
                println(pw, indent, "            }");
                println(pw, indent, "");
                println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
                println(pw, indent, "                throw new IllegalArgumentException(\"Wrong token: '\" + host + \"'\");");
                println(pw, indent, "            }");
                println(pw, indent, "        };");
            }
            else {
                println(pw, indent, "        return new ITokVisitor() {");
                println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
                String returnLine = "                return new " + name + "(";
                print(pw, indent, returnLine);
                pad = "";
                for (int i = 0; i < returnLine.length(); ++i ) {
                    pad = pad + " ";
                }
                boolean first = true;
                for (String ref: referencedSymbols) {
                    if (first) {
                        first = false;
                        pw.print("(" + ref + ") host");
                    }
                    else {
                        pw.println(",");
                        print(pw, indent, pad);
                        pw.print("(" + ref + ") nextToken()");
                    }
                    pw.print(".execute("+symbolToVariableName("parse"+ref)+", param)");
                }
                pw.println(");");
                println(pw, indent, "            }");
                println(pw, indent, "        };");
            }
        }
        println(pw, indent, "    }");
        println(pw, indent, "");
        println(pw, indent, "    /**");
        println(pw, indent, "     * Make a token visitor to parse an E non-terminal that decorates the given visitor.");
        println(pw, indent, "     *");
        println(pw, indent, "     * @param successor visitor to decorate");
        println(pw, indent, "     */");
        println(pw, indent, "    public ITokVisitor makeChainedVisitor(final ITokVisitor successor) {");
        if (def.isAlternative()) {
            print(pw, indent, "        return ");
            for (int i=0; i < referencedSymbols.size()-1; ++i) {
                String ref = referencedSymbols.get(i);
                pw.println(symbolToVariableName(ref+"Fact") + ".makeChainedVisitor(");
                print(pw, indent, "               ");
            }
            String ref = referencedSymbols.get(referencedSymbols.size()-1);
            pw.print(symbolToVariableName(ref+"Fact") + ".makeChainedVisitor(successor");
            for (int i=0; i < referencedSymbols.size(); ++i) {
                pw.print(")");
            }
            pw.println(";");
        }
        else {
            println(pw, indent, "        initialize();");
            if (g.isTerminal(origReferencedSymbols.get(0))) {
                // first one is terminal
                println(pw, indent, "        return new " + referencedSymbols.get(0) + ".I" + referencedSymbols.get(0) + "Visitor() {");
                println(pw, indent, "            public Object " + symbolToCaseName(referencedSymbols.get(0)) + "(" + referencedSymbols.get(0) + " host, Object param) {");
                println(pw, indent, "                return make" + name + "(host, param);");
                println(pw, indent, "            }");
                println(pw, indent, "");
                println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
                println(pw, indent, "                return host.execute(successor, param);");
                println(pw, indent, "            }");
                println(pw, indent, "        };");
            }
            else {
                println(pw, indent, "        return new ITokVisitor() {");
                println(pw, indent, "            public Object defaultCase(AToken host, Object param) {");
                String ref = referencedSymbols.get(0);
                println(pw, indent, "                Object o = host.execute(" + symbolToVariableName(ref+"Fact") + ".makeChainedVisitor(successor), param);");
                String returnLine = "                return (o instanceof " + ref + ") ? new " + name + "((" + ref + ") o";
                print(pw, indent, returnLine);
                pad = "";
                for (int i = 0; i < returnLine.length(); ++i ) {
                    pad = pad + " ";
                }
                boolean first = true;
                for (int refIndex = 1; refIndex<referencedSymbols.size(); ++ refIndex) {
                    ref = referencedSymbols.get(refIndex);
                    if (first) {
                        first = false;
                        pw.print(", ");
                        pw.print(" (" + ref + ") host");
                    }
                    else {
                        pw.println(",");
                        print(pw, indent, pad);
                        pw.print("(" + ref + ") nextToken()");
                    }
                    pw.print(".execute("+symbolToVariableName("parse"+ref)+", param)");
                }
                pw.println(") : o;");
                println(pw, indent, "            }");
                println(pw, indent, "        };");
            }
        }
        println(pw, indent, "    }");
        if (!def.isAlternative()) {
            println(pw, indent, "");
            println(pw, indent, "    /**");
            println(pw, indent, "     * Initialize this factory.");
            println(pw, indent, "     */");
            println(pw, indent, "    private void initialize() {");
            println(pw, indent, "        _initializer.apply(null);");
            println(pw, indent, "    }");
        }
    }

    /**
     * Get the name from a quoted string.
     * @param g grammar
     * @param symbol quoted string
     * @return name
     */
    private String quotedStringToName(Grammar g, String symbol) {
        String name;
        name = "_Terminal" + g.getTerminals().indexOf(symbol);
        return name;
    }

    /**
     * Get the case name from a symbol.
     * @param symbol
     * @return case name
     */
    protected String symbolToCaseName(String symbol) {
        int pos = (symbol.charAt(0)=='_'?2:1);
        return symbol.substring(0,pos).toLowerCase() + symbol.substring(pos)+"Case";
    }

    protected String symbolToVariableName(String symbol) {
        return "_" + symbol.toLowerCase();
    }
}
