package grammar.util;

import grammar.Grammar;
import grammar.Sequence;
import grammar.Definition;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.File;
import java.util.ArrayList;

/**
 * Generator for factories in inner classes.
 *
 * @author Mathias Ricken
 */
public class InnerClassesFactoryGenerator {
    /**
     * Singleton instance.
     */
    public static final InnerClassesFactoryGenerator Singleton = new InnerClassesFactoryGenerator();

    /**
     * Private singleton constructor.
     */
    private InnerClassesFactoryGenerator() {
        // nothing to do
    }


    private static void println(PrintWriter pw, int indent, String s) {
        for (int i=0; i<indent; ++i) {
            pw.print("    ");
        }
        pw.println(s);
    }

    private static void print(PrintWriter pw, int indent, String s) {
        for (int i=0; i<indent; ++i) {
            pw.print("    ");
        }
        pw.print(s);
    }

    public void generate(final String symbol, String path, String packageName, Grammar g) throws FileNotFoundException {
        String name = symbol;
        if (symbol.startsWith("\"")) {
            name = quotedStringToName(g, symbol);
        }

        PrintWriter pw = new PrintWriter(new FileOutputStream(path+File.separator+name+"Fact.java"), true);

        if (!packageName.equals("")) {
            pw.println("package "+packageName+";");
            pw.println();
        }

        ArrayList<String> symbols = new ArrayList<String>();
        symbols.add(symbol);

        generateHelper(pw, 0, symbols, path, packageName, g, new ILambda() {
            public Object apply(Object param) {
                return (param.equals(symbol)) ? "public" : "private";
            }
        });

        pw.close();
    }


    public void generateHelper(PrintWriter pw, int indent, ArrayList<String> symbols, String path, String packageName, Grammar g, ILambda accessSpecifier) throws FileNotFoundException {
        String symbol = symbols.get(symbols.size()-1);

        for (int i=0; i<indent; ++i) {
            System.out.print("  ");
        }
        System.out.println("Generating factory for " + symbol);

        pw.println("");

        if (symbol.equals("Empty")) {
            println(pw, indent, "");
            EmptyFactGenerator.Singleton.generate(pw, indent, path, packageName, g, accessSpecifier);
        }
        else {
            FactoryGenerator.Singleton.generate(pw, indent, symbol, path, packageName, g, false, symbols, accessSpecifier);
        }

        println(pw, indent, "");

        if (!g.isTerminal(symbol)) {
            Definition def = g.getDefinition(symbol);

            // this is not a terminal, process factories for referenced symbols
            if (def.isAlternative()) {
                for (Sequence seq: def.getAlternatives()) {
                    String sym = seq.getSymbols().get(0);
                    if (!symbols.contains(sym)) {
                        // this symbol has not been processed
                        ArrayList<String> newSymbols = new ArrayList<String>(symbols);
                        newSymbols.add(sym);
                        generateHelper(pw, indent+1, newSymbols, path, packageName, g, accessSpecifier);
                    }
                }
            }
            else {
                for (String sym: def.getAlternatives().get(0).getSymbols()) {
                    if (!symbols.contains(sym)) {
                        // this symbol has not been processed
                        ArrayList<String> newSymbols = new ArrayList<String>(symbols);
                        newSymbols.add(sym);
                        generateHelper(pw, indent+1, newSymbols, path, packageName, g, accessSpecifier);
                    }
                }

            }

        }

        println(pw, indent, "}");
    }

    /**
     * Get the name from a quoted string.
     * @param g grammar
     * @param symbol quoted string
     * @return name
     */
    private String quotedStringToName(Grammar g, String symbol) {
        String name;
        name = "_Terminal" + g.getTerminals().indexOf(symbol);
        return name;
    }
}
