package grammar.util;

import grammar.Grammar;
import grammar.Definition;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.File;
import java.util.ArrayList;

/**
 * Generator for terminals.
 *
 * @author Mathias Ricken
 */
public class TerminalGenerator {
    /**
     * Singleton instance.
     */
    public static final TerminalGenerator Singleton = new TerminalGenerator();

    /**
     * Private singleton constructor.
     */
    private TerminalGenerator() {
        // nothing to do
    }

    public void generate(String symbol, String path, String packageName, Grammar g) throws FileNotFoundException {
        String name = symbol;
        if (symbol.equals("Empty")) {
            EmptyGenerator.Singleton.generate(path, packageName, g);
            return;
        }

        if (symbol.startsWith("\"")) {
            name = "_Terminal" + g.getTerminals().indexOf(symbol);
        }

        PrintWriter pw = new PrintWriter(new FileOutputStream(path+File.separator+name+".java"), true);

        if (!packageName.equals("")) {
            pw.println("package "+packageName+";");
            pw.println();
        }
        pw.println("/**");
        pw.println(" * Terminal: " + name + ".");
        pw.println(" */");
        pw.print("public class " + name + " extends AToken");

        ArrayList<Definition> referers = g.getRefererDefinitions(name);
        boolean first = true;

        for (Definition d: referers) {
            if (d.isAlternative()) {
                if (first) {
                    pw.println();
                    pw.print("    implements " + d.getName());
                    first = false;
                }
                else {
                    pw.println(",");
                    pw.print("               " + d.getName());
                }
            }
        }

        pw.println(" {");

        pw.println("    /**");
        pw.println("     * Interface for a visitor to visit a " + name + ".");
        pw.println("     */");
        pw.println("    public static interface I" + name + "Visitor extends ITokVisitor {");
        pw.println("        /**");
        pw.println("         * Case for a " + name + ".");
        pw.println("         *");
        pw.println("         * @param host  " + name);
        pw.println("         * @param param visitor-specific parameter");
        pw.println("         * @return visitor-specific return value");
        pw.println("         */");
        pw.println("        public Object " + symbolToCaseName(name) + "(" + name + " host, Object param);");
        pw.println("    }");
        pw.println();
        if (symbol.startsWith("\"")) {
            pw.println("    /**");
            pw.println("     * Private singleton constructor for a " + name + ".");
            pw.println("     *");
            pw.println("     * @param lexeme lexeme");
            pw.println("     */");
            pw.println("    private " + name + "() {");
            pw.println("        super(" + symbol + ");");
            pw.println("    }");
            pw.println();
            pw.println("    /**");
            pw.println("     * Singleton instance.");
            pw.println("     */");
            pw.println("    public static final " + name + " Singleton = new " + name + "();");
        }
        else {
            pw.println("    /**");
            pw.println("     * Constructor for a " + name + ".");
            pw.println("     *");
            pw.println("     * @param lexeme lexeme");
            pw.println("     */");
            pw.println("    public " + name + "(String lexeme) {");
            pw.println("        super(lexeme);");
            pw.println("    }");
        }
        pw.println();
        pw.println("    /**");
        pw.println("     * Execute the " + name + " case of the visitor.");
        pw.println("     *");
        pw.println("     * @param algo  visitor");
        pw.println("     * @param param visitor-specific parameter");
        pw.println("     * @return visitor-specific return value");
        pw.println("     */");
        pw.println("    public Object execute(ITokVisitor algo, Object param) {");
        pw.println("        return (algo instanceof I" + name + "Visitor)");
        pw.println("                ? ((I" + name + "Visitor) algo)." + symbolToCaseName(name) + "(this, param)");
        pw.println("                : algo.defaultCase(this, param);");
        pw.println("    }");
        pw.println("}");

        pw.close();
    }

    /**
     * Get the case name from a symbol.
     * @param symbol
     * @return case name
     */
    protected String symbolToCaseName(String symbol) {
        int pos = (symbol.charAt(0)=='_'?2:1);
        return symbol.substring(0,pos).toLowerCase() + symbol.substring(pos)+"Case";
    }
}
