package parser;

/**
 * Factory for D1a grammar non-terminal objects.
 */
public class ParseD1aFactory extends ATVFactory {
    /**
     * Token visitor to parse D grammar non-terminals.
     */
    private IVisitDefault _parseD;

    /**
     * Constructor for the D1a factory.
     *
     * @param tkz  tokenizer to use
     * @param dFac factory for D grammar non-terminals
     */
    public ParseD1aFactory(ITokenizer tkz, ParseDFactory dFac) {
        super(tkz);
        _parseD = dFac.makeVisitor();
    }

    /**
     * Constructor for the D1a factory,
     * Does not take a factory for D grammar non-terminals. That factory has to be set using setParseTFactory().
     *
     * @param tkz tokenizer to use
     */
    public ParseD1aFactory(ITokenizer tkz) {
        super(tkz);
    }

    /**
     * Set the factory for D grammar non-terminals.
     *
     * @param dFac factory for D grammar non-terminals
     */
    public void setParseDFactory(ParseDFactory dFac) {
        _parseD = dFac.makeVisitor();
    }

    /**
     * Make an D1a grammar non-terminal.
     *
     * @param host semicolon token
     * @param inp  visitor-specific input
     * @return D1a grammar non-terminal
     */
    private D1a makeD1a(SemicolonToken host, Object inp) {
        return new D1a(host,
                       (D) nextToken().execute(_parseD, inp));
    }


    /**
     * Make a token visitor to parse an D1a non-terminal.
     *
     * @return token visitor
     */
    public IVisitDefault makeVisitor() {
        return new SemicolonToken.IVisitSemicolonToken() {
            public Object semicolonCase(SemicolonToken host, Object inp) {
                return makeD1a(host, inp);
            }

            public Object defaultCase(AToken host, Object param) {
                throw new IllegalArgumentException("Wrong token: '" + host + "'");
            }
        };
    }

    /**
     * Make a token visitor to parse an D1a non-terminal that decorates the given visitor.
     *
     * @param decoree visitor to decorate
     */
    public IVisitDefault makeDecVisitor(final IVisitDefault decoree) {
        return new SemicolonToken.IVisitSemicolonToken() {
            public Object semicolonCase(SemicolonToken host, Object inp) {
                return makeD1a(host, inp);
            }

            public Object defaultCase(AToken host, Object param) {
                return host.execute(decoree, param);
            }
        };
    }
}

