package parser;

/**
 * Factory for E1a grammar non-terminal objects.
 */
public class ParseE1aFactory extends ATVFactory {
    /**
     * Token visitor to parse E grammar non-terminals.
     */
    private IVisitDefault _parseE;

    /**
     * Constructor for the E1a factory.
     *
     * @param tkz  tokenizer to use
     * @param eFac factory for E grammar non-terminals
     */
    public ParseE1aFactory(ITokenizer tkz, ParseEFactory eFac) {
        super(tkz);
        _parseE = eFac.makeVisitor();
    }

    /**
     * Constructor for the E1a factory,
     * Does not take a factory for E grammar non-terminals. That factory has to be set using setParseEFactory().
     *
     * @param tkz tokenizer to use
     */
    public ParseE1aFactory(ITokenizer tkz) {
        super(tkz);
    }

    /**
     * Set the factory for E grammar non-terminals.
     *
     * @param eFac factory for E grammar non-terminals
     */
    public void setParseEFactory(ParseEFactory eFac) {
        _parseE = eFac.makeVisitor();
    }

    /**
     * Make an E1a grammar non-terminal.
     *
     * @param host or token
     * @param inp  visitor-specific input
     * @return E1a grammar non-terminal
     */
    private E1a makeE1a(OrToken host, Object inp) {
        return new E1a(host,
                       (E) nextToken().execute(_parseE, inp));
    }


    /**
     * Make a token visitor to parse an E1a non-terminal.
     *
     * @return token visitor
     */
    public IVisitDefault makeVisitor() {
        return new OrToken.IVisitOrToken() {
            public Object orCase(OrToken host, Object inp) {
                return makeE1a(host, inp);
            }

            public Object defaultCase(AToken host, Object param) {
                throw new IllegalArgumentException("Wrong token: '" + host + "'");
            }
        };
    }

    /**
     * Make a token visitor to parse an E1a non-terminal that decorates the given visitor.
     *
     * @param decoree visitor to decorate
     */
    public IVisitDefault makeDecVisitor(final IVisitDefault decoree) {
        return new OrToken.IVisitOrToken() {
            public Object orCase(OrToken host, Object inp) {
                return makeE1a(host, inp);
            }

            public Object defaultCase(AToken host, Object param) {
                return host.execute(decoree, param);
            }
        };
    }
}

