
package parser;

/**
 * Factory for T2 grammar non-terminal objects.
 */
public class ParseT2Factory extends ATVFactory {
    /**
     * Token visitor to parse T3 grammar non-terminals.
     */
    private IVisitDefault _parseT3;

    /**
     * Factory for QuotedString grammar terminals.
     */
    private IVisitDefault _parseQS;

    /**
     * Constructor for the T2 factory.
     *
     * @param tkz   tokenizer to use
     * @param t3Fac factory for T3 grammar non-terminals
     */
    public ParseT2Factory(ITokenizer tkz, ParseQuotedStringFactory qsFac, ParseT3Factory t3Fac) {
        super(tkz);
        _parseQS = qsFac.makeVisitor();
        _parseT3 = t3Fac.makeVisitor();
    }

    /**
     * Constructor for the T factory,
     * Does not take a factory for T3 grammar non-terminals. That factory has to be set using setParseT3Factory().
     *
     * @param tkz tokenizer to use
     */
    public ParseT2Factory(ITokenizer tkz, ParseQuotedStringFactory qsFac) {
        super(tkz);
        _parseQS = qsFac.makeVisitor();
    }

    /**
     * Set the factory for T3 grammar non-terminals.
     *
     * @param t3Fac factory for T3 grammar non-terminals
     */
    public void setParseT3Factory(ParseT3Factory t3Fac) {
        _parseT3 = t3Fac.makeVisitor();
    }

    /**
     * Make an T2 grammar non-terminal.
     *
     * @param host some token
     * @param inp  visitor-specific input
     * @return T grammar non-terminal
     */
    private T2 makeT(AToken host, Object inp) {
        return new T2((QuotedStringToken) host.execute(_parseQS, inp),
                     (T3) nextToken().execute(_parseT3, inp));
    }

    /**
     * Make a token visitor to parse an T non-terminal.
     *
     * @return token visitor
     */
    public IVisitDefault makeVisitor() {
        return new IVisitDefault() {
            public Object defaultCase(AToken host, Object param) {
                return makeT(host, param);
            }
        };
    }

    /**
     * Make a token visitor to parse an T non-terminal that decorates the given visitor.
     *
     * @param decoree visitor to decorate
     */
    public IVisitDefault makeDecVisitor(final IVisitDefault decoree) {
        return new QuotedStringToken.IVisitQuotedStringToken() {
            public Object quotedStringCase(QuotedStringToken host, Object param) {
                return makeT(host, param);
            }

            public Object defaultCase(AToken host, Object param) {
                return host.execute(decoree, param);
            }
        };
    }
}

